# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_i2c/i2c.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_i2c/i2c.proto',
  package='pw.i2c',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x10pw_i2c/i2c.proto\x12\x06pw.i2c\"e\n\x0fI2cWriteRequest\x12\x11\n\tbus_index\x18\x01 \x01(\r\x12\x16\n\x0etarget_address\x18\x02 \x01(\r\x12\x18\n\x10register_address\x18\x03 \x01(\x0c\x12\r\n\x05value\x18\x04 \x01(\x0c\"\x12\n\x10I2cWriteResponse\"h\n\x0eI2cReadRequest\x12\x11\n\tbus_index\x18\x01 \x01(\r\x12\x16\n\x0etarget_address\x18\x02 \x01(\r\x12\x18\n\x10register_address\x18\x03 \x01(\x0c\x12\x11\n\tread_size\x18\x04 \x01(\r\" \n\x0fI2cReadResponse\x12\r\n\x05value\x18\x01 \x01(\x0c\x32\x84\x01\n\x03I2c\x12?\n\x08I2cWrite\x12\x17.pw.i2c.I2cWriteRequest\x1a\x18.pw.i2c.I2cWriteResponse\"\x00\x12<\n\x07I2cRead\x12\x16.pw.i2c.I2cReadRequest\x1a\x17.pw.i2c.I2cReadResponse\"\x00\x62\x06proto3'
)




_I2CWRITEREQUEST = _descriptor.Descriptor(
  name='I2cWriteRequest',
  full_name='pw.i2c.I2cWriteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='bus_index', full_name='pw.i2c.I2cWriteRequest.bus_index', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target_address', full_name='pw.i2c.I2cWriteRequest.target_address', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='register_address', full_name='pw.i2c.I2cWriteRequest.register_address', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='pw.i2c.I2cWriteRequest.value', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=28,
  serialized_end=129,
)


_I2CWRITERESPONSE = _descriptor.Descriptor(
  name='I2cWriteResponse',
  full_name='pw.i2c.I2cWriteResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=131,
  serialized_end=149,
)


_I2CREADREQUEST = _descriptor.Descriptor(
  name='I2cReadRequest',
  full_name='pw.i2c.I2cReadRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='bus_index', full_name='pw.i2c.I2cReadRequest.bus_index', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target_address', full_name='pw.i2c.I2cReadRequest.target_address', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='register_address', full_name='pw.i2c.I2cReadRequest.register_address', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='read_size', full_name='pw.i2c.I2cReadRequest.read_size', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=151,
  serialized_end=255,
)


_I2CREADRESPONSE = _descriptor.Descriptor(
  name='I2cReadResponse',
  full_name='pw.i2c.I2cReadResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='pw.i2c.I2cReadResponse.value', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=257,
  serialized_end=289,
)

DESCRIPTOR.message_types_by_name['I2cWriteRequest'] = _I2CWRITEREQUEST
DESCRIPTOR.message_types_by_name['I2cWriteResponse'] = _I2CWRITERESPONSE
DESCRIPTOR.message_types_by_name['I2cReadRequest'] = _I2CREADREQUEST
DESCRIPTOR.message_types_by_name['I2cReadResponse'] = _I2CREADRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

I2cWriteRequest = _reflection.GeneratedProtocolMessageType('I2cWriteRequest', (_message.Message,), {
  'DESCRIPTOR' : _I2CWRITEREQUEST,
  '__module__' : 'pw_i2c.i2c_pb2'
  # @@protoc_insertion_point(class_scope:pw.i2c.I2cWriteRequest)
  })
_sym_db.RegisterMessage(I2cWriteRequest)

I2cWriteResponse = _reflection.GeneratedProtocolMessageType('I2cWriteResponse', (_message.Message,), {
  'DESCRIPTOR' : _I2CWRITERESPONSE,
  '__module__' : 'pw_i2c.i2c_pb2'
  # @@protoc_insertion_point(class_scope:pw.i2c.I2cWriteResponse)
  })
_sym_db.RegisterMessage(I2cWriteResponse)

I2cReadRequest = _reflection.GeneratedProtocolMessageType('I2cReadRequest', (_message.Message,), {
  'DESCRIPTOR' : _I2CREADREQUEST,
  '__module__' : 'pw_i2c.i2c_pb2'
  # @@protoc_insertion_point(class_scope:pw.i2c.I2cReadRequest)
  })
_sym_db.RegisterMessage(I2cReadRequest)

I2cReadResponse = _reflection.GeneratedProtocolMessageType('I2cReadResponse', (_message.Message,), {
  'DESCRIPTOR' : _I2CREADRESPONSE,
  '__module__' : 'pw_i2c.i2c_pb2'
  # @@protoc_insertion_point(class_scope:pw.i2c.I2cReadResponse)
  })
_sym_db.RegisterMessage(I2cReadResponse)



_I2C = _descriptor.ServiceDescriptor(
  name='I2c',
  full_name='pw.i2c.I2c',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=292,
  serialized_end=424,
  methods=[
  _descriptor.MethodDescriptor(
    name='I2cWrite',
    full_name='pw.i2c.I2c.I2cWrite',
    index=0,
    containing_service=None,
    input_type=_I2CWRITEREQUEST,
    output_type=_I2CWRITERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='I2cRead',
    full_name='pw.i2c.I2c.I2cRead',
    index=1,
    containing_service=None,
    input_type=_I2CREADREQUEST,
    output_type=_I2CREADRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_I2C)

DESCRIPTOR.services_by_name['I2c'] = _I2C

# @@protoc_insertion_point(module_scope)
