# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_digital_io/digital_io.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_digital_io/digital_io.proto',
  package='pw.digital_io',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x1epw_digital_io/digital_io.proto\x12\rpw.digital_io\"<\n\x16\x44igitalIoEnableRequest\x12\x12\n\nline_index\x18\x01 \x01(\r\x12\x0e\n\x06\x65nable\x18\x02 \x01(\x08\"\x19\n\x17\x44igitalIoEnableResponse\"\\\n\x18\x44igitalIoSetStateRequest\x12\x12\n\nline_index\x18\x01 \x01(\r\x12,\n\x05state\x18\x02 \x01(\x0e\x32\x1d.pw.digital_io.DigitalIoState\"\x1b\n\x19\x44igitalIoSetStateResponse\".\n\x18\x44igitalIoGetStateRequest\x12\x12\n\nline_index\x18\x01 \x01(\r\"I\n\x19\x44igitalIoGetStateResponse\x12,\n\x05state\x18\x01 \x01(\x0e\x32\x1d.pw.digital_io.DigitalIoState**\n\x0e\x44igitalIoState\x12\x0c\n\x08INACTIVE\x10\x00\x12\n\n\x06\x41\x43TIVE\x10\x01\x32\xa8\x02\n\tDigitalIo\x12Y\n\x06\x45nable\x12%.pw.digital_io.DigitalIoEnableRequest\x1a&.pw.digital_io.DigitalIoEnableResponse\"\x00\x12_\n\x08SetState\x12\'.pw.digital_io.DigitalIoSetStateRequest\x1a(.pw.digital_io.DigitalIoSetStateResponse\"\x00\x12_\n\x08GetState\x12\'.pw.digital_io.DigitalIoGetStateRequest\x1a(.pw.digital_io.DigitalIoGetStateResponse\"\x00\x62\x06proto3'
)

_DIGITALIOSTATE = _descriptor.EnumDescriptor(
  name='DigitalIoState',
  full_name='pw.digital_io.DigitalIoState',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INACTIVE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ACTIVE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=384,
  serialized_end=426,
)
_sym_db.RegisterEnumDescriptor(_DIGITALIOSTATE)

DigitalIoState = enum_type_wrapper.EnumTypeWrapper(_DIGITALIOSTATE)
INACTIVE = 0
ACTIVE = 1



_DIGITALIOENABLEREQUEST = _descriptor.Descriptor(
  name='DigitalIoEnableRequest',
  full_name='pw.digital_io.DigitalIoEnableRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='line_index', full_name='pw.digital_io.DigitalIoEnableRequest.line_index', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='enable', full_name='pw.digital_io.DigitalIoEnableRequest.enable', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=49,
  serialized_end=109,
)


_DIGITALIOENABLERESPONSE = _descriptor.Descriptor(
  name='DigitalIoEnableResponse',
  full_name='pw.digital_io.DigitalIoEnableResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=111,
  serialized_end=136,
)


_DIGITALIOSETSTATEREQUEST = _descriptor.Descriptor(
  name='DigitalIoSetStateRequest',
  full_name='pw.digital_io.DigitalIoSetStateRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='line_index', full_name='pw.digital_io.DigitalIoSetStateRequest.line_index', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='pw.digital_io.DigitalIoSetStateRequest.state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=138,
  serialized_end=230,
)


_DIGITALIOSETSTATERESPONSE = _descriptor.Descriptor(
  name='DigitalIoSetStateResponse',
  full_name='pw.digital_io.DigitalIoSetStateResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=232,
  serialized_end=259,
)


_DIGITALIOGETSTATEREQUEST = _descriptor.Descriptor(
  name='DigitalIoGetStateRequest',
  full_name='pw.digital_io.DigitalIoGetStateRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='line_index', full_name='pw.digital_io.DigitalIoGetStateRequest.line_index', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=261,
  serialized_end=307,
)


_DIGITALIOGETSTATERESPONSE = _descriptor.Descriptor(
  name='DigitalIoGetStateResponse',
  full_name='pw.digital_io.DigitalIoGetStateResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='state', full_name='pw.digital_io.DigitalIoGetStateResponse.state', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=309,
  serialized_end=382,
)

_DIGITALIOSETSTATEREQUEST.fields_by_name['state'].enum_type = _DIGITALIOSTATE
_DIGITALIOGETSTATERESPONSE.fields_by_name['state'].enum_type = _DIGITALIOSTATE
DESCRIPTOR.message_types_by_name['DigitalIoEnableRequest'] = _DIGITALIOENABLEREQUEST
DESCRIPTOR.message_types_by_name['DigitalIoEnableResponse'] = _DIGITALIOENABLERESPONSE
DESCRIPTOR.message_types_by_name['DigitalIoSetStateRequest'] = _DIGITALIOSETSTATEREQUEST
DESCRIPTOR.message_types_by_name['DigitalIoSetStateResponse'] = _DIGITALIOSETSTATERESPONSE
DESCRIPTOR.message_types_by_name['DigitalIoGetStateRequest'] = _DIGITALIOGETSTATEREQUEST
DESCRIPTOR.message_types_by_name['DigitalIoGetStateResponse'] = _DIGITALIOGETSTATERESPONSE
DESCRIPTOR.enum_types_by_name['DigitalIoState'] = _DIGITALIOSTATE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DigitalIoEnableRequest = _reflection.GeneratedProtocolMessageType('DigitalIoEnableRequest', (_message.Message,), {
  'DESCRIPTOR' : _DIGITALIOENABLEREQUEST,
  '__module__' : 'pw_digital_io.digital_io_pb2'
  # @@protoc_insertion_point(class_scope:pw.digital_io.DigitalIoEnableRequest)
  })
_sym_db.RegisterMessage(DigitalIoEnableRequest)

DigitalIoEnableResponse = _reflection.GeneratedProtocolMessageType('DigitalIoEnableResponse', (_message.Message,), {
  'DESCRIPTOR' : _DIGITALIOENABLERESPONSE,
  '__module__' : 'pw_digital_io.digital_io_pb2'
  # @@protoc_insertion_point(class_scope:pw.digital_io.DigitalIoEnableResponse)
  })
_sym_db.RegisterMessage(DigitalIoEnableResponse)

DigitalIoSetStateRequest = _reflection.GeneratedProtocolMessageType('DigitalIoSetStateRequest', (_message.Message,), {
  'DESCRIPTOR' : _DIGITALIOSETSTATEREQUEST,
  '__module__' : 'pw_digital_io.digital_io_pb2'
  # @@protoc_insertion_point(class_scope:pw.digital_io.DigitalIoSetStateRequest)
  })
_sym_db.RegisterMessage(DigitalIoSetStateRequest)

DigitalIoSetStateResponse = _reflection.GeneratedProtocolMessageType('DigitalIoSetStateResponse', (_message.Message,), {
  'DESCRIPTOR' : _DIGITALIOSETSTATERESPONSE,
  '__module__' : 'pw_digital_io.digital_io_pb2'
  # @@protoc_insertion_point(class_scope:pw.digital_io.DigitalIoSetStateResponse)
  })
_sym_db.RegisterMessage(DigitalIoSetStateResponse)

DigitalIoGetStateRequest = _reflection.GeneratedProtocolMessageType('DigitalIoGetStateRequest', (_message.Message,), {
  'DESCRIPTOR' : _DIGITALIOGETSTATEREQUEST,
  '__module__' : 'pw_digital_io.digital_io_pb2'
  # @@protoc_insertion_point(class_scope:pw.digital_io.DigitalIoGetStateRequest)
  })
_sym_db.RegisterMessage(DigitalIoGetStateRequest)

DigitalIoGetStateResponse = _reflection.GeneratedProtocolMessageType('DigitalIoGetStateResponse', (_message.Message,), {
  'DESCRIPTOR' : _DIGITALIOGETSTATERESPONSE,
  '__module__' : 'pw_digital_io.digital_io_pb2'
  # @@protoc_insertion_point(class_scope:pw.digital_io.DigitalIoGetStateResponse)
  })
_sym_db.RegisterMessage(DigitalIoGetStateResponse)



_DIGITALIO = _descriptor.ServiceDescriptor(
  name='DigitalIo',
  full_name='pw.digital_io.DigitalIo',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=429,
  serialized_end=725,
  methods=[
  _descriptor.MethodDescriptor(
    name='Enable',
    full_name='pw.digital_io.DigitalIo.Enable',
    index=0,
    containing_service=None,
    input_type=_DIGITALIOENABLEREQUEST,
    output_type=_DIGITALIOENABLERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SetState',
    full_name='pw.digital_io.DigitalIo.SetState',
    index=1,
    containing_service=None,
    input_type=_DIGITALIOSETSTATEREQUEST,
    output_type=_DIGITALIOSETSTATERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetState',
    full_name='pw.digital_io.DigitalIo.GetState',
    index=2,
    containing_service=None,
    input_type=_DIGITALIOGETSTATEREQUEST,
    output_type=_DIGITALIOGETSTATERESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_DIGITALIO)

DESCRIPTOR.services_by_name['DigitalIo'] = _DIGITALIO

# @@protoc_insertion_point(module_scope)
